# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of h5config.
#
# h5config is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# h5config is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# h5config.  If not, see <http://www.gnu.org/licenses/>.

import os as _os
import os.path as _os_path
import sys as _sys
import types as _types


class Storage (object):
    "A storage bakend for loading and saving `Config` instances"
    def load(self, config, merge=False, **kwargs):
        if merge:
            self.clear()
        self._load(config=config, **kwargs)
        config.set_storage(storage=self)

    def _load(self, config, **kwargs):
        raise NotImplementedError()

    def save(self, config, merge=False, **kwargs):
        if merge:
            self.clear()
        self._save(config=config, **kwargs)
        config._storage = self

    def _save(self, config, **kwargs):
        raise NotImplementedError()

    def clear(self):
        raise NotImplementedError()


class FileStorage (Storage):
    "`Config` storage backend by a single file"
    extension = None

    def __init__(self, filename=None):
        self._filename = filename

    def _create_basedir(self, filename):
        dirname = _os_path.dirname(filename)
        if dirname and not _os_path.isdir(dirname):
            _os.makedirs(dirname)


def is_string(x):
    if _sys.version_info >= (3,):
        return isinstance(x, (bytes, str))
    else:  # Python 2 compatibility
        return isinstance(x, _types.StringTypes)



# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of h5config.
#
# h5config is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# h5config is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# h5config.  If not, see <http://www.gnu.org/licenses/>.

"""Tools for setting up a package using config files.

The benefit of subclassing `PackageConfig` over using something like
`configparser` is that you can easily store default `h5config` values
in the configuration file.  Consider the following example:

TODO
> class _MyConfig
"""

import logging as _logging
import os.path as _os_path
import sys as _sys

from . import config as _config
from . import log as _log
from .storage.hdf5 import HDF5_Storage as _HDF5_Storage
from .storage.yaml import YAML_Storage as _YAML_Storage


class PackageConfig (_config.Config):
    """Configure package operation

    This basic implementation just creates and manages a package-wide
    `LOG` instance.  If you create this instance on your own (for
    example, to work around bootstrapping issues), just pass your
    instance in as `logger` when you initialize this class.
    """
    possible_storage = [_HDF5_Storage, _YAML_Storage]
    settings = [
        _config.ChoiceSetting(
            name='log-level',
            help='Module logging level.',
            default=_logging.WARN,
            choices=[
                ('critical', _logging.CRITICAL),
                ('error', _logging.ERROR),
                ('warn', _logging.WARN),
                ('info', _logging.INFO),
                ('debug', _logging.DEBUG),
                ]),
        _config.BooleanSetting(
            name='syslog',
            help='Log to syslog (otherwise log to stderr).',
            default=False),
        ]

    def __init__(self, package_name, namespace=None, logger=None, **kwargs):
        super(PackageConfig, self).__init__(**kwargs)
        self._package_name = package_name
        if not namespace:
            namespace = _sys.modules[package_name]
        self._namespace = namespace
        if not logger:
            logger = _log.get_basic_logger(package_name, level=_logging.WARN)
        self._logger = logger
        if 'LOG' not in dir(namespace):
            namespace.LOG = logger

    def setup(self):
        self._logger.setLevel(self['log-level'])
        if self['syslog']:
            if 'syslog' not in self._logger._handler_cache:
                _syslog_handler = _logging_handlers.SysLogHandler()
                _syslog_handler.setLevel(_logging.DEBUG)
                self._logger._handler_cache['syslog'] = _syslog_handler
                self._logger.handlers = [self._logger._handler_cache['syslog']]
        else:
            self._logger.handlers = [self._logger._handler_cache['stream']]
        self._logger.info('setup {} packge config:\n{}'.format(
                self._package_name, self.dump()))

    def clear(self):
        "Replace self with a non-backed version with default settings."
        super(PackageConfig, self).clear()
        self._storage = None

    def _base_paths(self):
        user_basepath = _os_path.join(
            _os_path.expanduser('~'), '.config', self._package_name)
        system_basepath = _os_path.join('/etc', self._package_name, 'config')
        distributed_basepath =  _os_path.join(
            '/usr', 'share', self._package_name, 'config')
        return [user_basepath, system_basepath, distributed_basepath]

    def load_system(self):
        "Return the best `PackageConfig` match after scanning the filesystem"
        self._logger.info('looking for package config file')
        basepaths = self._base_paths()
        for basepath in basepaths:
            for storage in self.possible_storage:
                filename = '{}.{}'.format(basepath, storage.extension)
                if _os_path.exists(filename):
                    self._logger.info(
                        'base_config file found at {}'.format(filename))
                    self._storage = storage(filename=filename)
                    self.load()
                    self.setup()
                    return
                else:
                    self._logger.debug(
                        'no base_config file at {}'.format(filename))
        # create (but don't save) the preferred file
        basepath = basepaths[0]
        storage = self.possible_storage[0]
        filename = '{}.{}'.format(basepath, storage.extension)
        self._logger.info('new base_config file at {}'.format(filename))
        self._storage = storage(filename=filename)
        self.load()
        self.setup()



# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of h5config.
#
# h5config is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# h5config is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# h5config.  If not, see <http://www.gnu.org/licenses/>.

"""The basic h5config classes
"""

import copy as _copy

from . import LOG as _LOG


class Setting (object):
    "A named setting with arbitrary text values."
    def  __init__(self, name, help='', default=None):
        self.name = name
        self._help = help
        self.default = default

    def __str__(self):
        return '<%s %s>' % (self.__class__.__name__, self.name)

    def __repr__(self):
        return self.__str__()

    def help(self):
        ret = '%s  Default: %s.' % (
            self._help, self.convert_to_text(self.default))
        return ret.strip()

    def convert_from_text(self, value):
        return value or None

    def convert_to_text(self, value):
        return value or ''


class ChoiceSetting (Setting):
    """A named setting with a limited number of possible values.

    `choices` should be a list of `(config_file_value, Python value)`
    pairs.  For example

    >>> s = ChoiceSetting(name='bool',
    ...                   choices=[('yes', True), ('no', False)])
    >>> s.convert_from_text('yes')
    True
    >>> s.convert_to_text(True)
    'yes'
    >>> s.convert_to_text('invalid')
    Traceback (most recent call last):
      ...
    ValueError: invalid
    >>> s.help()
    'Default: yes.  Choices: yes, no'
    """
    def __init__(self, choices=None, **kwargs):
        if 'default' not in kwargs:
            if None not in [keyval[1] for keyval in choices]:
                kwargs['default'] = choices[0][1]
        super(ChoiceSetting, self).__init__(**kwargs)
        if choices == None:
            choices = []
        self.choices = choices

    def help(self):
        ret = '%s  Choices: %s' % (
            super(ChoiceSetting, self).help(),
            ', '.join([key for key,value in self.choices]))
        return ret.strip()

    def convert_from_text(self, value):
        return dict(self.choices)[value]

    def convert_to_text(self, value):
        for keyval in self.choices:
            key,val = keyval
            if val == value:
                return key
        raise ValueError(value)


class BooleanSetting (ChoiceSetting):
    """A named settubg that can be either true or false.

    >>> s = BooleanSetting(name='bool')

    >>> s.convert_from_text('yes')
    True
    >>> s.convert_to_text(True)
    'yes'
    >>> s.convert_to_text('invalid')
    Traceback (most recent call last):
      ...
    ValueError: invalid
    >>> s.help()
    'Default: no.  Choices: yes, no'
    """
    def __init__(self, default=False, **kwargs):
        super(BooleanSetting, self).__init__(
            choices=[('yes', True), ('no', False)], default=default, **kwargs)


class NumericSetting (Setting):
    """A named setting with numeric values.

    Don't use this setting class.  Use a more specific subclass, such
    as `IntegerSetting`.

    >>> s = NumericSetting(name='float')
    >>> s.default
    0
    >>> s.convert_to_text(13)
    '13'
    """
    def __init__(self, default=0, **kwargs):
        super(NumericSetting, self).__init__(default=default, **kwargs)

    def convert_to_text(self, value):
        return str(value)

    def convert_from_text(self, value):
        if value in [None, 'None']:
            return None
        return self._convert_from_text(value)

    def _convert_from_text(self, value):
        raise NotImplementedError()


class IntegerSetting (NumericSetting):
    """A named setting with integer values.

    >>> s = IntegerSetting(name='int')
    >>> s.default
    1
    >>> s.convert_from_text('8')
    8
    """
    def __init__(self, default=1, **kwargs):
        super(IntegerSetting, self).__init__(default=default, **kwargs)

    def _convert_from_text(self, value):
        return int(value)


class FloatSetting (NumericSetting):
    """A named setting with floating point values.

    >>> s = FloatSetting(name='float')
    >>> s.default
    1.0
    >>> s.convert_from_text('8')
    8.0
    >>> try:
    ...     s.convert_from_text('invalid')
    ... except ValueError as e:
    ...     print('caught a ValueError')
    caught a ValueError
    """
    def __init__(self, default=1.0, **kwargs):
        super(FloatSetting, self).__init__(default=default, **kwargs)

    def _convert_from_text(self, value):
        return float(value)


class ListSetting(Setting):
    """A named setting with a list of string values.

    >>> s = ListSetting(name='list')
    >>> s.default
    []
    >>> s.convert_to_text(['abc', 'def'])
    'abc,def'
    >>> s.convert_from_text('uvw, xyz')
    ['uvw', ' xyz']
    >>> s.convert_to_text([])
    ''
    >>> s.convert_from_text('')
    []
    """
    def __init__(self, default=None, **kwargs):
        if default is None:
            default = []
        super(ListSetting, self).__init__(default=default, **kwargs)

    def _convert_from_text(self, value):
        return value

    def convert_from_text(self, value):
        if value is None:
            return None
        elif value in ['', []]:
            return []
        return [self._convert_from_text(x) for x in value.split(',')]

    def convert_to_text(self, value):
        if value is None:
            return None
        return ','.join([str(x) for x in value])

class IntegerListSetting (ListSetting):
    """A named setting with a list of integer point values.

    >>> s = IntegerListSetting(name='integerlist')
    >>> s.default
    []
    >>> s.convert_to_text([1, 3])
    '1,3'
    >>> s.convert_from_text('4, -6')
    [4, -6]
    >>> s.convert_to_text([])
    ''
    >>> s.convert_from_text('')
    []
    """
    def _convert_from_text(self, value):
        if value is None:
            return value
        return int(value)


class FloatListSetting (ListSetting):
    """A named setting with a list of floating point values.

    >>> s = FloatListSetting(name='floatlist')
    >>> s.default
    []
    >>> s.convert_to_text([1, 2.3])
    '1,2.3'
    >>> s.convert_from_text('4.5, -6.7')  # doctest: +ELLIPSIS
    [4.5, -6.7...]
    >>> s.convert_to_text([])
    ''
    >>> s.convert_from_text('')
    []
    """
    def _convert_from_text(self, value):
        if value is None:
            return value
        return float(value)


class ConfigSetting (Setting):
    """A setting that holds a pointer to a child `Config` class

    This allows you to nest `Config`\s, which is a useful way to
    contain complexity.  In order to save such a config, the backend
    must be able to handle hierarchical storage (possibly via
    references).

    For example, a configurable AFM may contain a configurable piezo
    scanner, as well as a configurable stepper motor.
    """
    def __init__(self, config_class=None, **kwargs):
        super(ConfigSetting, self).__init__(**kwargs)
        self.config_class = config_class


class ConfigListSetting (ConfigSetting):
    """A setting that holds a list of child `Config` classes

    For example, a piezo scanner with several axes.
    """
    def __init__(self, default=[], **kwargs):
        super(ConfigListSetting, self).__init__(**kwargs)


class Config (dict):
    """A class with a list of `Setting`\s.

    `Config` instances store their values just like dictionaries, but
    the attached list of `Settings`\s allows them to intelligently
    dump, save, and load those values.
    """
    settings = []

    def __init__(self, storage=None):
        super(Config, self).__init__()
        self.clear()
        self.set_storage(storage=storage)

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, id(self))

    def set_storage(self, storage):
        self._storage = storage

    def clear(self):
        super(Config, self).clear()
        for s in self.settings:
            # copy to avoid ambiguity with mutable defaults
            self[s.name] = _copy.deepcopy(s.default)

    def load(self, merge=False, **kwargs):
        self._storage.load(config=self, merge=merge, **kwargs)

    def save(self, merge=False, **kwargs):
        self._storage.save(config=self, merge=merge, **kwargs)

    def dump(self, help=False, prefix=''):
        """Return all settings and their values as a string

        >>> class MyConfig (Config):
        ...     settings = [
        ...         ChoiceSetting(
        ...             name='number',
        ...             help='I have a number behind my back...',
        ...             default=1,
        ...             choices=[('one', 1), ('two', 2),
        ...                 ]),
        ...         BooleanSetting(
        ...             name='odd',
        ...             help='The number behind my back is odd.',
        ...             default=True),
        ...         IntegerSetting(
        ...             name='guesses',
        ...             help='Number of guesses before epic failure.',
        ...             default=2),
        ...         ]
        >>> c = MyConfig()
        >>> print(c.dump())
        number: one
        odd: yes
        guesses: 2
        >>> print(c.dump(help=True))  # doctest: +NORMALIZE_WHITESPACE
        number: one (I have a number behind my back...  Default: one.
                     Choices: one, two)
        odd: yes    (The number behind my back is odd.  Default: yes.
                     Choices: yes, no)
        guesses: 2  (Number of guesses before epic failure.  Default: 2.)
        """
        lines = []
        for setting in self.settings:
            name = setting.name
            value = self[name]
            try:
                if isinstance(setting, ConfigListSetting):
                    if value:
                        lines.append('{}{}:'.format(prefix, name))
                        for i,config in enumerate(value):
                            lines.append('{}  {}:'.format(prefix, i))
                            lines.append(
                                config.dump(help=help, prefix=prefix+'    '))
                        continue
                elif isinstance(setting, ConfigSetting):
                    if value is not None:
                        lines.append('{}{}:'.format(prefix, name))
                        lines.append(value.dump(help=help, prefix=prefix+'  '))
                        continue
                value_string = setting.convert_to_text(self[name])
                if help:
                    help_string = '\t({})'.format(setting.help())
                else:
                    help_string = ''
                lines.append('{}{}: {}{}'.format(
                        prefix, name, value_string, help_string))
            except Exception:
                _LOG.error('could not dump {} ({!r})'.format(name, value))
                raise
        return '\n'.join(lines)

    def select_config(self, setting_name, attribute_value=None,
                      get_attribute=lambda value : value['name']):
        """Select a `Config` instance from `ConfigListSetting` values

        If your don't want to select `ConfigListSetting` items by
        index, you can select them by matching another attribute.  For
        example:

        >>> from .test import TestConfig
        >>> c = TestConfig()
        >>> c['children'] = []
        >>> for name in ['Jack', 'Jill']:
        ...     child = TestConfig()
        ...     child['name'] = name
        ...     c['children'].append(child)
        >>> child = c.select_config('children', 'Jack')
        >>> child  # doctest: +ELLIPSIS
        <TestConfig ...>
        >>> child['name']
        'Jack'

        `get_attribute` defaults to returning `value['name']`, because
        I expect that to be the most common case, but you can use
        another function if neccessary, for example to drill down more
        than one level.  Here we select by grandchild name:

        >>> for name in ['Jack', 'Jill']:
        ...     grandchild = TestConfig()
        ...     grandchild['name'] = name + ' Junior'
        ...     child = c.select_config('children', name)
        ...     child['children'] = [grandchild]
        >>> get_grandchild = lambda value : value['children'][0]['name']
        >>> get_grandchild(child)
        'Jill Junior'
        >>> child = c.select_config('children', 'Jack Junior',
        ...     get_attribute=get_grandchild)
        >>> child  # doctest: +ELLIPSIS
        <TestConfig ...>
        >>> child['name']
        'Jack'
        """
        setting_value = self.get(setting_name, None)
        if not setting_value:
            raise KeyError(setting_name)
        for item in setting_value:
            if get_attribute(item) == attribute_value:
                return item
        raise KeyError(attribute_value)



# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of h5config.
#
# h5config is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# h5config is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# h5config.  If not, see <http://www.gnu.org/licenses/>.

from .log import get_basic_logger as _get_basic_logger


__version__ = '0.2'

LOG = _get_basic_logger(__name__)

from .tools import PackageConfig as _PackageConfig


package_config = _PackageConfig(package_name=__name__, logger=LOG)
package_config.load_system()



# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of h5config.
#
# h5config is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# h5config is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# h5config.  If not, see <http://www.gnu.org/licenses/>.

"""HDF5 backend implementation
"""

from __future__ import absolute_import

import os.path as _os_path

import yaml as _yaml  # global PyYAML module

from .. import LOG as _LOG
from .. import config as _config
from . import FileStorage as _FileStorage
from . import is_string as _is_string


class _YAMLDumper (_yaml.SafeDumper):
    def represent_bool(self, data):
        "Use yes/no instead of the default true/false"
        if data:
            value = 'yes'
        else:
            value = 'no'
        return self.represent_scalar('tag:yaml.org,2002:bool', value)


_YAMLDumper.add_representer(bool, _YAMLDumper.represent_bool)


class YAML_Storage (_FileStorage):
    """Back a `Config` class with a YAML file.

    >>> import os
    >>> from ..test import TestConfig
    >>> import os.path
    >>> import tempfile
    >>> fd,filename = tempfile.mkstemp(
    ...     suffix='.'+YAML_Storage.extension, prefix='pypiezo-')
    >>> os.close(fd)

    >>> c = TestConfig(storage=YAML_Storage(filename=filename))
    >>> c.load()

    Saving writes all the config values to disk.

    >>> c['alive'] = True
    >>> c.save()
    >>> print(open(filename, 'r').read())  # doctest: +REPORT_UDIFF
    age: 1.3
    alive: yes
    bids:
    - 5.4
    - 3.2
    - 1
    children: ''
    claws:
    - 1
    - 2
    daisies: 13
    name: ''
    species: Norwegian Blue
    spouse: ''
    words:
    - cracker
    - wants
    <BLANKLINE>

    Loading reads the config files from disk.

    >>> c.clear()
    >>> c['alive']
    False
    >>> c.load()
    >>> c['alive']
    True

    Cleanup our temporary config file.

    >>> os.remove(filename)
    """
    extension = 'yaml'
    dumper = _YAMLDumper

    def _load(self, config):
        if not _os_path.exists(self._filename):
            open(self._filename, 'a').close()
        with open(self._filename, 'r') as f:
            data = _yaml.safe_load(f)
        if data == None:
            return  # empty file
        return self._from_dict(config, data)

    @staticmethod
    def _from_dict(config, data):
        settings = dict([(s.name, s) for s in config.settings])
        for key,value in data.items():
            setting = settings[key]
            if isinstance(setting, (_config.BooleanSetting,
                                    _config.NumericSetting,
                                    _config.ListSetting,
                                    _config.IntegerListSetting,
                                    _config.FloatListSetting)):
                if _is_string(value):
                    # older versions of h5config
                    value = s.convert_from_text(value)
                v = value
            elif isinstance(setting, _config.ConfigListSetting) and value:
                values = []
                for v in value:
                    values.append(YAML_Storage._from_dict(
                            setting.config_class(), v))
                v = values
            elif isinstance(setting, _config.ConfigSetting) and value:
                v = YAML_Storage._from_dict(setting.config_class(), value)
            else:
                v = setting.convert_from_text(value)
            config[key] = v
        return config

    def _save(self, config):
        self._create_basedir(filename=self._filename)
        data = self._to_dict(config)
        with open(self._filename, 'w') as f:
            _yaml.dump(data, stream=f, Dumper=self.dumper,
                       default_flow_style=False)

    @staticmethod
    def _to_dict(config):
        data = {}
        settings = dict([(s.name, s) for s in config.settings])
        for key,value in config.items():
            if key in settings:
                setting = settings[key]
                if isinstance(setting, (_config.BooleanSetting,
                                        _config.NumericSetting,
                                        _config.ListSetting,
                                        _config.IntegerListSetting,
                                        _config.FloatListSetting)):
                    v = value
                elif isinstance(setting, _config.ConfigListSetting) and value:
                    values = []
                    for v in value:
                        values.append(YAML_Storage._to_dict(v))
                    v = values
                elif isinstance(setting, _config.ConfigSetting) and value:
                    v = YAML_Storage._to_dict(value)
                else:
                    v = setting.convert_to_text(value)
                data[key] = v
        return data



# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of h5config.
#
# h5config is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# h5config is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# h5config.  If not, see <http://www.gnu.org/licenses/>.

"""Define a test config object using all the setting types

The first time you a storage backend, the file it creates will
probably be empty or not exist.

>>> import os
>>> import tempfile
>>> from h5config.storage.hdf5 import pprint_HDF5, HDF5_Storage

>>> fd,filename = tempfile.mkstemp(suffix='.h5', prefix='h5config-')
>>> os.close(fd)

>>> c = TestConfig(storage=HDF5_Storage(filename=filename, group='/base'))
>>> c.load()

Loading will create a stub group group if it hadn't existed before.

>>> pprint_HDF5(filename)
/
  /base

Saving fills in all the config values.

>>> c['syslog'] = True
>>> c.save()
>>> pprint_HDF5(filename)  # doctest: +REPORT_UDIFF, +ELLIPSIS
/
  /base
    <HDF5 dataset "age": shape (), type "<f8">
      1.3
    <HDF5 dataset "alive": shape (), type "|b1">
      False
    <HDF5 dataset "bids": shape (3,), type "<f8">
      [ 5.4  3.2  1. ]
    <HDF5 dataset "children": shape (), type "|S1">
<BLANKLINE>
    <HDF5 dataset "claws": shape (2,), type "<i8">
      [1 2]
    <HDF5 dataset "daisies": shape (), type "<i...">
      13
    <HDF5 dataset "name": shape (), type "|S1">
<BLANKLINE>
    <HDF5 dataset "species": shape (), type "|S14">
      Norwegian Blue
    <HDF5 dataset "spouse": shape (), type "|S1">
<BLANKLINE>
    <HDF5 dataset "words": shape (2,), type "|S7">
      ['cracker' 'wants']

If you want more details, you can dump with help strings.

>>> print(c.dump(help=True))  # doctest: +REPORT_UDIFF, +NORMALIZE_WHITESPACE
name:                    (The parrot's name.  Default: .)
species: Norwegian Blue  (Type of parrot.  Default: Norwegian Blue.
                          Choices: Norwegian Blue, Macaw)
alive: no                (The parrot is alive.  Default: no.  Choices: yes, no)
daisies: 13              (Number of daisies pushed up by the parrot.
                          Default: 13.)
age: 1.3                 (Parrot age in years  Default: 1.3.)
words: cracker,wants     (Words known by the parrot.  Default: cracker,wants.)
claws: 1,2               (Claws on each foot.  Default: 1,2.)
bids: 5.4,3.2,1          (Prices offered for parrot.  Default: 5.4,3.2,1.)
spouse:                  (This parrot's significant other.  Default: .)
children:                (This parrot's children.  Default: .)

As you can see from the `age` setting, settings also support `None`,
even if they have numeric types.

Cleanup our temporary config file.

>>> os.remove(filename)
"""

import os as _os
import sys as _sys
import tempfile as _tempfile

from . import LOG as _LOG
from . import config as _config
from .storage import FileStorage as _FileStorage
from .storage.hdf5 import HDF5_Storage
from .storage.yaml import YAML_Storage


class TestConfig (_config.Config):
    "Test all the setting types for the h5config module"
    settings = [
        _config.Setting(
            name='name',
            help="The parrot's name."),
        _config.ChoiceSetting(
            name='species',
            help='Type of parrot.',
            default=0,
            choices=[('Norwegian Blue', 0), ('Macaw', 1)]),
        _config.BooleanSetting(
            name='alive',
            help='The parrot is alive.',
            default=False),
        _config.IntegerSetting(
            name='daisies',
            help="Number of daisies pushed up by the parrot.",
            default=13),
        _config.FloatSetting(
            name='age',
            help='Parrot age in years',
            default=1.3),
        _config.ListSetting(
            name='words',
            help='Words known by the parrot.',
            default=['cracker', 'wants']),
        _config.IntegerListSetting(
            name='claws',
            help='Claws on each foot.',
            default=[1, 2]),
        _config.FloatListSetting(
            name='bids',
            help='Prices offered for parrot.',
            default=[5.4, 3.2, 1]),
        _config.ConfigSetting(
            name='spouse',
            help="This parrot's significant other."),
        _config.ConfigListSetting(
            name='children',
            help="This parrot's children."),
        ]

# must define self-references after completing the TestConfig class
for s in TestConfig.settings:
    if s.name in ['spouse', 'children']:
        s.config_class = TestConfig


def _alternative_test_config(name):
    ret = TestConfig()
    ret['name'] = name
    return ret

_ALTERNATIVES = {  # alternative settings for testing
    'name': 'Captain Flint',
    'species': 1,
    'alive': True,
    'daisies': None,
    'age': None,
    'words': ['arrrr', 'matey'],
    'claws': [3, 0],
    'bids': [],
    'spouse': _alternative_test_config(name='Lory'),
    'children': [_alternative_test_config(name=n)
                 for n in ['Washington Post', 'Eli Yale']],
    }
# TODO: share children with spouse to test references

def test(storage=None):
    if storage is None:
        storage = [HDF5_Storage, YAML_Storage]
        for s in storage:
            test(storage=s)
        return
    _LOG.debug('testing {}'.format(storage))
    _basic_tests(storage)
    if issubclass(storage, _FileStorage):
        _file_storage_tests(storage)

def _basic_tests(storage):
    pass

def _file_storage_tests(storage):
    fd,filename = _tempfile.mkstemp(
        suffix='.'+storage.extension, prefix='h5config-')
    _os.close(fd)
    try:
        c = TestConfig(storage=storage(filename=filename))
        c.dump()
        c.save()
        c.load()
        nd = list(_non_defaults(c))
        assert not nd, (storage, nd)
        for key,value in _ALTERNATIVES.items():
            c[key] = value
        c.dump()
        c.save()
        na = dict(_non_alternatives(c))
        assert not na, (storage, na)
        c.clear()
        nd = list(_non_defaults(c))
        assert not nd, (storage, nd)
        c.load()
        na = dict(_non_alternatives(c))
        assert not na, (storage, na)
    finally:
        _os.remove(filename)

def _non_defaults(config):
    for setting in TestConfig.settings:
        value = config[setting.name]
        if value != setting.default:
            yield (setting.name, value)

def _non_alternatives(config, alternatives=None):
    if alternatives is None:
        alternatives = _ALTERNATIVES
    for setting in TestConfig.settings:
        value = config[setting.name]
        alt = alternatives[setting.name]
        if value != alt:
            _LOG.error('{} value missmatch: {} vs {}'.format(
                    setting.name, value, alt))
            yield (setting.name, value)
        elif type(value) != type(alt):
            _LOG.error('{} type missmatch: {} vs {}'.format(
                    setting.name, type(value), type(alt)))
            yield (setting.name, value)



# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of h5config.
#
# h5config is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# h5config is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# h5config.  If not, see <http://www.gnu.org/licenses/>.

"Conveniently save and load config-options from HDF5 and YAML files."

from distutils.core import setup
import os.path

from h5config import __version__


package_name = 'h5config'
_this_dir = os.path.dirname(__file__)

setup(name=package_name,
      version=__version__,
      maintainer='W. Trevor King',
      maintainer_email='wking@tremily.us',
      url='http://blog.tremily.us/posts/{}/'.format(package_name),
      download_url='http://git.tremily.us/?p={}.git;a=snapshot;h=v{};sf=tgz'.format(
        package_name, __version__),
      license='GNU General Public License (GPL)',
      platforms=['all'],
      description=__doc__,
      long_description=open(os.path.join(_this_dir, 'README'), 'r').read(),
      classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
      packages=[package_name, '{}.storage'.format(package_name)],
      provides=['{} ({})'.format(package_name, __version__)],
      )



# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of h5config.
#
# h5config is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# h5config is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# h5config.  If not, see <http://www.gnu.org/licenses/>.

"""Tools for setting up a package logging.

This module is separate from `tools` to avoid module dependency
cycles.  This module has no internal dependencies, while `tools`
depends on many of the other modules.  With this module separate, the
other internal modules have access to the default logger before the
package configuration is built up enough to configure it according to
your external specifications.
"""

import logging as _logging


def get_basic_logger(name, level=_logging.WARN):
    """Create and return a basic logger

    This utility function encapsulates a bunch of `logging`
    boilerplate that I use in several packages.
    """
    log = _logging.getLogger(name)
    log.setLevel(level)
    formatter = _logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    stream_handler = _logging.StreamHandler()
    stream_handler.setLevel(_logging.DEBUG)
    stream_handler.setFormatter(formatter)
    log.addHandler(stream_handler)
    # Cache handlers for easy swapping depending on config settings
    log._handler_cache = {'stream': stream_handler}
    return log



# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of h5config.
#
# h5config is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# h5config is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# h5config.  If not, see <http://www.gnu.org/licenses/>.

"""HDF5 backend implementation
"""

import os.path as _os_path
import sys as _sys

import h5py as _h5py
import numpy as _numpy

from .. import LOG as _LOG
from .. import config as _config
from . import FileStorage as _FileStorage
from . import is_string as _is_string


def pprint_HDF5(*args, **kwargs):
    print(pformat_HDF5(*args, **kwargs))

def pformat_HDF5(filename, group='/'):
    try:
        with _h5py.File(filename, 'r') as f:
            cwg = f[group]
            ret = '\n'.join(_pformat_hdf5(cwg))
    except IOError as e:
        if 'unable to open' in e.message:
            if _os_path.getsize(filename) == 0:
                return 'EMPTY'
            return None
        raise
    return ret

def _pformat_hdf5(cwg, depth=0):
    lines = []
    lines.append('  '*depth + cwg.name)
    depth += 1
    for key,value in cwg.items():
        if isinstance(value, _h5py.Group):
            lines.extend(_pformat_hdf5(value, depth))
        elif isinstance(value, _h5py.Dataset):
            lines.append('  '*depth + str(value))
            lines.append('  '*(depth+1) + str(value[...]))
        else:
            lines.append('  '*depth + str(value))
    return lines

def h5_create_group(cwg, path, force=False):
    "Create the group where the settings are stored (if necessary)."
    if path == '/':
        return cwg
    gpath = ['']
    for group in path.strip('/').split('/'):
        gpath.append(group)
        if group not in cwg.keys():
            _LOG.debug('creating group {} in {}'.format(
                    '/'.join(gpath), cwg.file))
            cwg.create_group(group)
        _cwg = cwg[group]
        if isinstance(_cwg, _h5py.Dataset):
            if force:
                _LOG.info('overwrite {} in {} ({}) with a group'.format(
                        '/'.join(gpath), _cwg.file, _cwg))
                del cwg[group]
                _cwg = cwg.create_group(group)
            else:
                raise ValueError(_cwg)
        cwg = _cwg
    return cwg


class HDF5_Storage (_FileStorage):
    """Back a `Config` class with an HDF5 file.

    The `.save` and `.load` methods have an optional `group` argument
    that allows you to save and load settings from an externally
    opened HDF5 file.  This can make it easier to stash several
    related `Config` classes in a single file.  For example

    >>> import os
    >>> import tempfile
    >>> from ..test import TestConfig
    >>> fd,filename = tempfile.mkstemp(
    ...     suffix='.'+HDF5_Storage.extension, prefix='pypiezo-')
    >>> os.close(fd)

    >>> f = _h5py.File(filename, 'a')
    >>> c = TestConfig(storage=HDF5_Storage(
    ...     filename='untouched_file.h5', group='/untouched/group'))
    >>> c['alive'] = True
    >>> group = f.create_group('base')
    >>> c.save(group=group)
    >>> pprint_HDF5(filename)  # doctest: +REPORT_UDIFF, +ELLIPSIS
    /
      /base
        <HDF5 dataset "age": shape (), type "<f8">
          1.3
        <HDF5 dataset "alive": shape (), type "|b1">
          True
        <HDF5 dataset "bids": shape (3,), type "<f8">
          [ 5.4  3.2  1. ]
        <HDF5 dataset "children": shape (), type "|S1">
    <BLANKLINE>
        <HDF5 dataset "claws": shape (2,), type "<i8">
          [1 2]
        <HDF5 dataset "daisies": shape (), type "<i...">
          13
        <HDF5 dataset "name": shape (), type "|S1">
    <BLANKLINE>
        <HDF5 dataset "species": shape (), type "|S14">
          Norwegian Blue
        <HDF5 dataset "spouse": shape (), type "|S1">
    <BLANKLINE>
        <HDF5 dataset "words": shape (2,), type "|S7">
          ['cracker' 'wants']
    >>> c.clear()
    >>> c['alive']
    False
    >>> c.load(group=group)
    >>> c['alive']
    True

    >>> f.close()
    >>> os.remove(filename)
    """
    extension = 'h5'

    def __init__(self, group='/', **kwargs):
        super(HDF5_Storage, self).__init__(**kwargs)
        if isinstance(group, _h5py.Group):
            self._file_checked = True
        else:
            assert group.startswith('/'), group
            if not group.endswith('/'):
                group += '/'
            self._file_checked = False
        self.group = group

    def _check_file(self):
        if self._file_checked:
            return
        self._setup_file()
        self._file_checked = True

    def _setup_file(self):
        self._create_basedir(filename=self._filename)
        with _h5py.File(self._filename, 'a') as f:
            cwg = f  # current working group
            h5_create_group(cwg, self.group)

    def _load(self, config, group=None):
        f = None
        try:
            if group is None:
                if isinstance(self.group, _h5py.Group):
                    group = self.group
                else:
                    self._check_file()
                    f = _h5py.File(self._filename, 'r')
                    group = f[self.group]
            for s in config.settings:
                if s.name not in group.keys():
                    continue
                if isinstance(s, _config.ConfigListSetting):
                    try:
                        cwg = h5_create_group(group, s.name)
                    except ValueError:
                        pass
                    else:
                        value = []
                        for i in sorted(int(x) for x in cwg.keys()):
                            instance = s.config_class()
                            try:
                                _cwg = h5_create_group(cwg, str(i))
                            except ValueError:
                                pass
                            else:
                                self._load(config=instance, group=_cwg)
                                value.append(instance)
                        config[s.name] = value
                elif isinstance(s, _config.ConfigSetting):
                    try:
                        cwg = h5_create_group(group, s.name)
                    except ValueError:
                        pass
                    else:
                        if not config[s.name]:
                            config[s.name] = s.config_class()
                        self._load(config=config[s.name], group=cwg)
                else:
                    try:
                        v = group[s.name][...]
                    except Exception as e:
                        _LOG.error('Could not access {}/{}: {}'.format(
                                group.name, s.name, e))
                        raise 
                    if isinstance(v, _numpy.ndarray):
                        if isinstance(s, _config.BooleanSetting):
                            v = bool(v)  # array(True, dtype=bool) -> True
                        elif v.dtype.type == _numpy.string_:
                            if isinstance(s, _config.ListSetting):
                                try:
                                    v = list(v)
                                except TypeError:
                                    v = []
                                if _sys.version_info >= (3,):
                                    for i,v_ in enumerate(v):
                                        if isinstance(v_, bytes):
                                            v[i] = str(v_, 'utf-8')
                            else:  # array('abc', dtype='|S3') -> 'abc'
                                if _sys.version_info >= (3,):
                                    v = str(v, 'utf-8')
                                else:
                                    v = str(v)
                        elif isinstance(s, _config.IntegerSetting):
                            v = int(v)  # array(3, dtpe='int32') -> 3
                        elif isinstance(s, _config.FloatSetting):
                            v = float(v)  # array(1.2, dtype='float64') -> 1.2
                        elif isinstance(s, _config.NumericSetting):
                            raise NotImplementedError(type(s))
                        elif isinstance(s, _config.ListSetting):
                            v = list(v)  # convert from numpy array
                    if _is_string(v):
                        # convert back from None, etc.
                        v = s.convert_from_text(v)
                    config[s.name] = v
        finally:
            if f:
                f.close()

    def _save(self, config, group=None):
        f = None
        try:
            if group is None:
                if isinstance(self.group, _h5py.Group):
                    group = self.group
                else:
                    self._check_file()
                    f = _h5py.File(self._filename, 'a')
                    group = f[self.group]
            for s in config.settings:
                value = None
                if isinstance(s, (_config.BooleanSetting,
                                  _config.NumericSetting,
                                  _config.ListSetting)):
                    value = config[s.name]
                    if value in [None, []]:
                        value = s.convert_to_text(value)
                elif isinstance(s, _config.ConfigListSetting):
                    configs = config[s.name]
                    if configs:
                        cwg = h5_create_group(group, s.name, force=True)
                        for i,cfg in enumerate(configs):
                            _cwg = h5_create_group(cwg, str(i), force=True)
                            self._save(config=cfg, group=_cwg)
                        continue
                elif isinstance(s, _config.ConfigSetting):
                    cfg = config[s.name]
                    if cfg:
                        cwg = h5_create_group(group, s.name, force=True)
                        self._save(config=cfg, group=cwg)
                        continue
                if value is None:  # not set yet, or invalid
                    value = s.convert_to_text(config[s.name])
                if _sys.version_info >= (3,):  # convert strings to bytes/
                    if isinstance(value, str):
                        value = value.encode('utf-8')
                    elif isinstance(value, list):
                        value = list(value)  # shallow copy
                        for i,v in enumerate(value):
                            if isinstance(v, str):
                                value[i] = v.encode('utf-8')
                try:
                    del group[s.name]
                except KeyError:
                    pass
                try:
                    group[s.name] = value
                except TypeError:
                    raise ValueError((value, type(value)))
        finally:
            if f:
                f.close()
